import os
import subprocess
import sys
from types import ModuleType
from typing import TYPE_CHECKING, List, Optional, Union

import psutil

if TYPE_CHECKING:
    try:
        import ensight
    except ImportError:
        from ansys.api.pyensight import ensight_api


class Omniverse:
    """Provides the ``ensight.utils.omniverse`` interface.

    The omniverse class methods provide an interface between an EnSight session
    and an Omniverse instance.  See :ref:`omniverse_info` for additional details.

    Parameters
    ----------
    interface:
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.

    Notes
    -----
    This interface is only available when using pyensight (they do not work with
    the ensight Python interpreter) and the module must be used in an interpreter
    that includes the Omniverse Python modules (e.g. omni and pxr).  Only a single
    Omniverse connection can be established within a single pyensight session.

    Examples
    --------
    ::
        from ansys.pyensight.core import LocalLauncher
        session = LocalLauncher().start()
        ov = session.ensight.utils.omniverse
        ov.create_connection()
        ov.update()
        ov.close_connection()

    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface
        self._server_pid: Optional[int] = None
        self._interpreter: List[str] = []

    def _check_modules(self) -> None:
        """Verify that the Python interpreter is correct

        Check for omni module. If not present, raise an exception.
        If pxr is there as well, then we can just use sys.executable.
        If not, check to see if 'kit.bat' or 'kit.sh' can be found and
        arrange to use those instead.

        Raises
        ------
        RuntimeError if the necessary modules are missing.

        """
        # One time check for this
        if len(self._interpreter):
            return
        try:
            # Note: the EnSight embedded interpreter will not have these
            import omni.client  # noqa: F401
        except ImportError:
            raise RuntimeError("The module requires the omni module to be installed.") from None

        try:
            # if we can import pxr, then we can just use sys.executable
            from pxr import Gf, Sdf, Usd, UsdGeom, UsdLux, UsdShade  # noqa: F401

            if os.path.basename(sys.executable).startswith("kit"):
                # we are running inside of an Omniverse app like Create, use the 'kit' script
                raise ImportError("Internal retry")

            self._interpreter = [sys.executable]
            return
        except ImportError:
            # Can we find 'kit.bat' or 'kit.sh' (we may be running in it)?
            # Interesting cases:  something/kit/python/python.exe,
            # something/kit/kit.exe.  All mapped to something/kit.{bat,sh} if found.
            ov_dir = os.path.dirname(sys.executable)
            for _ in range(3):
                for name in ("kit.bat", "kit.sh"):
                    exe_name = os.path.join(ov_dir, name)
                    if os.path.exists(exe_name):
                        self._interpreter = [
                            exe_name,
                            "--enable",
                            "omni.client",
                            "--enable",
                            "omni.usd",
                            "--exec",
                        ]
                        return
                ov_dir = os.path.dirname(ov_dir)
            raise RuntimeError("Unable to detect a copy of the Omniverse kit executable.") from None

    def _is_running_omniverse(self) -> bool:
        """Check that an Omniverse connection is active
        Returns
        -------
            True if the connection is active, False otherwise.
        """
        if self._server_pid is None:
            return False
        if psutil.pid_exists(self._server_pid):
            return True
        self._server_pid = None
        return False

    def create_connection(
        self,
        omniverse_path: str,
        include_camera: bool = False,
        normalize_geometry: bool = False,
        live: bool = True,
        debug_filename: str = "",
    ) -> None:
        """Ensure that an EnSight dsg -> omniverse server is running

        Connect the current EnSight session to an Omniverse server.
        This is done by launching a new service that makes a dynamic scene graph
        connection to the EnSight session and pushes updates to the Omniverse server.
        The initial EnSight scene will be pushed after the connection is established.

        Parameters
        ----------
        omniverse_path : str
            The URI to the Omniverse server. It will look like this:
            "omniverse://localhost/Users/test"
        include_camera : bool
            If True, apply the EnSight camera to the Omniverse scene.  This option
            should be used if the target viewer is in AR/VR mode.  Defaults to False.
        normalize_geometry : bool
            Omniverse units are in meters.  If the source dataset is not in the correct
            unit system or is just too large/small, this option will remap the geometry
            to a unit cube.  Defaults to False.
        live : bool
            If True, one can call 'update()' to send updated geometry to Omniverse.
            If False, the Omniverse connection will push a single update and then
            disconnect.  Defaults to True.
        debug_filename : str
            If the name of a file is provided, it will be used to save logging information on
            the connection between EnSight and Omniverse.

        """
        if not isinstance(self._ensight, ModuleType):
            self._ensight._session.ensight_version_check("2023 R2")
        self._check_modules()
        if self._is_running_omniverse():
            raise RuntimeError("An Omniverse server connection is already active.")
        # Make sure the internal ui module is loaded
        self._ensight._session.cmd("import enspyqtgui_int", do_eval=False)
        # Get the gRPC connection details and use them to launch the service
        port = self._ensight._session.grpc.port()
        hostname = self._ensight._session.grpc.host
        token = self._ensight._session.grpc.security_token
        script_name = "omniverse_dsg_server.py"
        working_dir = os.path.dirname(__file__)
        cmd = [
            script_name,
            "--host",
            hostname,
            "--port",
            str(port),
            "--path",
            omniverse_path,
        ]
        if live:
            cmd.extend(["--live"])
        if include_camera:
            cmd.extend(["--vrmode"])
        if token:
            cmd.extend(["--security", token])
        # if temporal:
        #     cmd.extend(["--animation"])
        # else:
        #     cmd.extend(["--no-animation"])
        if debug_filename:
            cmd.extend(["--log_file", debug_filename])
            cmd.extend(["--verbose", "1"])
        if normalize_geometry:
            cmd.extend(["--normalize_geometry"])
        # if using kit.bat, convert args into a string, otherwise, just use them
        cmdline = []
        cmdline.extend(self._interpreter)
        if len(self._interpreter) > 1:
            cmd = [" ".join(cmd)]
        cmdline.extend(cmd)
        env_vars = os.environ.copy()
        process = subprocess.Popen(cmdline, close_fds=True, env=env_vars, cwd=working_dir)
        self._server_pid = process.pid

    def close_connection(self) -> None:
        """Shut down the open EnSight dsg -> omniverse server

        Break the connection between the EnSight instance and Omniverse.

        """
        self._check_modules()
        if not self._is_running_omniverse():
            return
        proc = psutil.Process(self._server_pid)
        for child in proc.children(recursive=True):
            if psutil.pid_exists(child.pid):
                # This can be a race condition, so it is ok if the child is dead already
                try:
                    child.kill()
                except psutil.NoSuchProcess:
                    pass
        # Same issue, this process might already be shutting down, so NoSuchProcess is ok.
        try:
            proc.kill()
        except psutil.NoSuchProcess:
            pass
        self._server_pid = None

    def update(self) -> None:
        """Update the geometry in Omniverse

        Push the current EnSight scene to the current Omniverse connection.

        """
        if not isinstance(self._ensight, ModuleType):
            self._ensight._session.ensight_version_check("2023 R2")
        self._check_modules()
        if not self._is_running_omniverse():
            raise RuntimeError("No Omniverse server connection is currently active.")
        update_cmd = "dynamicscenegraph://localhost/client/update"
        cmd = f'enspyqtgui_int.dynamic_scene_graph_command("{update_cmd}")'
        self._ensight._session.cmd(cmd, do_eval=False)



import logging
import os
import queue
import threading
from typing import Any, Dict, List, Optional

from ansys.api.pyensight.v0 import dynamic_scene_graph_pb2
from ansys.pyensight.core import ensight_grpc
import numpy


class Part(object):
    def __init__(self, session: "DSGSession"):
        """
        This object roughly represents an EnSight "Part".  It contains the connectivity,
        coordinates, normals and texture coordinate information for one DSG entity

        This object stores basic geometry information coming from the DSG protocol.  The
        update_geom() method can parse an "UpdateGeom" protobuffer and merges the results
        into the Part object.

        Parameters
        ----------
        session:
            The DSG connection session object.
        """
        self.session = session
        self.conn_tris = numpy.array([], dtype="int32")
        self.conn_lines = numpy.array([], dtype="int32")
        self.coords = numpy.array([], dtype="float32")
        self.normals = numpy.array([], dtype="float32")
        self.normals_elem = False
        self.tcoords = numpy.array([], dtype="float32")
        self.tcoords_var_id: Optional[int] = None
        self.tcoords_elem = False
        self.cmd: Optional[Any] = None
        self.reset()

    def reset(self, cmd: Any = None) -> None:
        self.conn_tris = numpy.array([], dtype="int32")
        self.conn_lines = numpy.array([], dtype="int32")
        self.coords = numpy.array([], dtype="float32")
        self.normals = numpy.array([], dtype="float32")
        self.normals_elem = False
        self.tcoords = numpy.array([], dtype="float32")
        self.tcoords_var_id = None
        self.tcoords_elem = False
        self.cmd = cmd

    def update_geom(self, cmd: dynamic_scene_graph_pb2.UpdateGeom) -> None:
        """
        Merge an update geometry command into the numpy buffers being cached in this object

        Parameters
        ----------
        cmd:
            This is an array update command.  It could be for coordinates, normals, variables, connectivity, etc.
        """
        if cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.COORDINATES:
            if self.coords.size != cmd.total_array_size:
                self.coords = numpy.resize(self.coords, cmd.total_array_size)
            self.coords[cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)] = cmd.flt_array
        elif cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.TRIANGLES:
            if self.conn_tris.size != cmd.total_array_size:
                self.conn_tris = numpy.resize(self.conn_tris, cmd.total_array_size)
            self.conn_tris[cmd.chunk_offset : cmd.chunk_offset + len(cmd.int_array)] = cmd.int_array
        elif cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.LINES:
            if self.conn_lines.size != cmd.total_array_size:
                self.conn_lines = numpy.resize(self.conn_lines, cmd.total_array_size)
            self.conn_lines[
                cmd.chunk_offset : cmd.chunk_offset + len(cmd.int_array)
            ] = cmd.int_array
        elif (cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_NORMALS) or (
            cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.NODE_NORMALS
        ):
            self.normals_elem = cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_NORMALS
            if self.normals.size != cmd.total_array_size:
                self.normals = numpy.resize(self.normals, cmd.total_array_size)
            self.normals[cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)] = cmd.flt_array
        elif (cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_VARIABLE) or (
            cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.NODE_VARIABLE
        ):
            # Get the variable definition
            if cmd.variable_id in self.session.variables:
                self.tcoords_var_id = cmd.variable_id
                self.tcoords_elem = (
                    cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_VARIABLE
                )
                if self.tcoords.size != cmd.total_array_size:
                    self.tcoords = numpy.resize(self.tcoords, cmd.total_array_size)
                self.tcoords[
                    cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)
                ] = cmd.flt_array
            else:
                self.tcoords_var_id = None

    def build(self):
        """
        This function processes the geometry arrays and converts them into nodal representation.
        It will duplicate triangles as needed (to preserve element normals) and will convert
        variable data into texture coordinates.

        Returns
        -------
        On failure, the method returns None for the first return value.  The returned tuple is:

        (part_command, vertices, connectivity, normals, tex_coords, var_command)

        part_command: UPDATE_PART command object
        vertices: numpy array of the nodal coordinates
        connectivity: numpy array of the triangle indices into the vertices array
        normals: numpy array of per vertex normal values (optional)
        tcoords: numpy array of per vertex texture coordinates (optional)
        var_command: UPDATE_VARIABLE command object for the variable the texture coordinate correspond to, if any
        """
        if self.cmd is None:
            return None, None, None, None, None, None
        if self.conn_lines.size:
            self.session.log(
                f"Note: part '{self.cmd.name}' contains lines which are not currently supported."
            )
            return None, None, None, None, None, None
        verts = self.coords
        if self.session.normalize_geometry and self.session.scene_bounds is not None:
            midx = (self.session.scene_bounds[3] + self.session.scene_bounds[0]) * 0.5
            midy = (self.session.scene_bounds[4] + self.session.scene_bounds[1]) * 0.5
            midz = (self.session.scene_bounds[5] + self.session.scene_bounds[2]) * 0.5
            dx = self.session.scene_bounds[3] - self.session.scene_bounds[0]
            dy = self.session.scene_bounds[4] - self.session.scene_bounds[1]
            dz = self.session.scene_bounds[5] - self.session.scene_bounds[2]
            s = dx
            if dy > s:
                s = dy
            if dz > s:
                s = dz
            if s == 0:
                s = 1.0
            num_verts = int(verts.size / 3)
            for i in range(num_verts):
                j = i * 3
                verts[j + 0] = (verts[j + 0] - midx) / s
                verts[j + 1] = (verts[j + 1] - midy) / s
                verts[j + 2] = (verts[j + 2] - midz) / s

        conn = self.conn_tris
        normals = self.normals
        tcoords = None
        if self.tcoords.size:
            tcoords = self.tcoords
        if self.tcoords_elem or self.normals_elem:
            verts_per_prim = 3
            num_prims = int(conn.size / verts_per_prim)
            # "flatten" the triangles to move values from elements to nodes
            new_verts = numpy.ndarray((num_prims * verts_per_prim * 3,), dtype="float32")
            new_conn = numpy.ndarray((num_prims * verts_per_prim,), dtype="int32")
            new_tcoords = None
            if tcoords is not None:
                # remember that the input values are 1D at this point, we will expand to 2D later
                new_tcoords = numpy.ndarray((num_prims * verts_per_prim,), dtype="float32")
            new_normals = None
            if normals is not None:
                if normals.size == 0:
                    self.session.log("Warning: zero length normals!")
                else:
                    new_normals = numpy.ndarray((num_prims * verts_per_prim * 3,), dtype="float32")
            j = 0
            for i0 in range(num_prims):
                for i1 in range(verts_per_prim):
                    idx = conn[i0 * verts_per_prim + i1]
                    # new connectivity (identity)
                    new_conn[j] = j
                    # copy the vertex
                    new_verts[j * 3 + 0] = verts[idx * 3 + 0]
                    new_verts[j * 3 + 1] = verts[idx * 3 + 1]
                    new_verts[j * 3 + 2] = verts[idx * 3 + 2]
                    if new_normals is not None:
                        if self.normals_elem:
                            # copy the normal associated with the face
                            new_normals[j * 3 + 0] = normals[i0 * 3 + 0]
                            new_normals[j * 3 + 1] = normals[i0 * 3 + 1]
                            new_normals[j * 3 + 2] = normals[i0 * 3 + 2]
                        else:
                            # copy the same normal as the vertex
                            new_normals[j * 3 + 0] = normals[idx * 3 + 0]
                            new_normals[j * 3 + 1] = normals[idx * 3 + 1]
                            new_normals[j * 3 + 2] = normals[idx * 3 + 2]
                    if new_tcoords is not None:
                        # remember, 1D texture coords at this point
                        if self.tcoords_elem:
                            # copy the texture coord associated with the face
                            new_tcoords[j] = tcoords[i0]
                        else:
                            # copy the same texture coord as the vertex
                            new_tcoords[j] = tcoords[idx]
                    j += 1
            # new arrays.
            verts = new_verts
            conn = new_conn
            normals = new_normals
            if tcoords is not None:
                tcoords = new_tcoords

        var_cmd = None
        # texture coords need transformation from variable value to [ST]
        if tcoords is not None:
            var_dsg_id = self.cmd.color_variableid
            var_cmd = self.session.variables[var_dsg_id]
            v_min = None
            v_max = None
            for lvl in var_cmd.levels:
                if (v_min is None) or (v_min > lvl.value):
                    v_min = lvl.value
                if (v_max is None) or (v_max < lvl.value):
                    v_max = lvl.value
            var_minmax = [v_min, v_max]
            # build a power of two x 1 texture
            num_texels = int(len(var_cmd.texture) / 4)
            half_texel = 1 / (num_texels * 2.0)
            num_verts = int(verts.size / 3)
            tmp = numpy.ndarray((num_verts * 2,), dtype="float32")
            tmp.fill(0.5)  # fill in the T coordinate...
            tex_width = half_texel * 2 * (num_texels - 1)  # center to center of num_texels
            # if the range is 0, adjust the min by -1.   The result is that the texture
            # coords will get mapped to S=1.0 which is what EnSight does in this situation
            if (var_minmax[1] - var_minmax[0]) == 0.0:
                var_minmax[0] = var_minmax[0] - 1.0
            var_width = var_minmax[1] - var_minmax[0]
            for idx in range(num_verts):
                # normalized S coord value (clamp)
                s = (tcoords[idx] - var_minmax[0]) / var_width
                if s < 0.0:
                    s = 0.0
                if s > 1.0:
                    s = 1.0
                # map to the texture range and set the S value
                tmp[idx * 2] = s * tex_width + half_texel
            tcoords = tmp

        self.session.log(
            f"Part '{self.cmd.name}' defined: {self.coords.size/3} verts, {self.conn_tris.size/3} tris, {self.conn_lines.size/2} lines."
        )
        command = self.cmd

        return command, verts, conn, normals, tcoords, var_cmd


class UpdateHandler(object):
    """
    This class serves as the interface between a DSGSession and a hosting application.
    The DSGSession processes the general aspects of the gRPC pipeline and collects the
    various DSG objects into collections of: groups, variables, etc.  It also coalesces
    the individual array updates into a "Part" object which represents a single addressable
    mesh chunk.
    UpdateHandler methods are called as the various update happen, and it is called when
    a mesh chunk has been entirely defined.  In most scenarios, a subclass of UpdateHandler
    is passed to the DSGSession to handshake the mesh data to the application target.
    """

    def __init__(self) -> None:
        self._session: "DSGSession"

    @property
    def session(self) -> "DSGSession":
        """The session object this handler has been associated with"""
        return self._session

    @session.setter
    def session(self, session: "DSGSession") -> None:
        self._session = session

    def add_group(self, id: int, view: bool = False) -> None:
        """Called when a new group command has been added: self.session.groups[id]"""
        if view:
            self.session.log(f"Adding view: {self.session.groups[id]}")
        else:
            self.session.log(f"Adding group: {self.session.groups[id].name}")

    def add_variable(self, id: int) -> None:
        """Called when a new group command has been added: self.session.variables[id]"""
        self.session.log(f"Adding variable: {self.session.variables[id].name}")

    def finalize_part(self, part: Part) -> None:
        """Called when all the updates on a Part object have been completed.

        Note: this superclass method should be called after the subclass has processed
        the part geometry as the saved part command will be destroyed by this call.
        """
        if part.cmd:
            self.session.log(f"Part finalized: {part.cmd.name}")
        part.cmd = None

    def start_connection(self) -> None:
        """A new gRPC connection has been established:  self.session.grpc"""
        grpc = self.session.grpc
        self.session.log(f"gRPC connection established to: {grpc.host}:{grpc.port}")

    def end_connection(self) -> None:
        """The previous gRPC connection has been closed"""
        self.session.log("gRPC connection closed")

    def begin_update(self) -> None:
        """A new scene update is about to begin"""
        self.session.log("Begin update ------------------------")

    def end_update(self) -> None:
        """The scene update is complete"""
        self.session.log("End update ------------------------")

    def get_dsg_cmd_attribute(self, obj: Any, name: str, default: Any = None) -> Optional[str]:
        """Utility function to get an attribute from a DSG update object

        Note: UpdateVariable and UpdateGroup commands support generic attributes
        """
        return obj.attributes.get(name, default)

    def group_matrix(self, group: Any) -> Any:
        matrix = group.matrix4x4
        # The Case matrix is basically the camera transform.  In vrmode, we only want
        # the raw geometry, so use the identity matrix.
        if (
            self.get_dsg_cmd_attribute(group, "ENS_OBJ_TYPE") == "ENS_CASE"
        ) and self.session.vrmode:
            matrix = [
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
            ]
        return matrix


class DSGSession(object):
    def __init__(
        self,
        port: int = 12345,
        host: str = "127.0.0.1",
        security_code: str = "",
        verbose: int = 0,
        normalize_geometry: bool = False,
        vrmode: bool = False,
        handler: UpdateHandler = UpdateHandler(),
    ):
        """
        Manage a gRPC connection and link it to an UpdateHandler instance

        This class makes a DSG gRPC connection via the specified port and host (leveraging
        the passed security code).  As DSG protobuffers arrive, they are merged into Part
        object instances and the UpdateHandler is invoked to further process them.

        Parameters
        ----------
        port : int
            The port number the EnSight gRPC service is running on.
            The default is ``12345``.
        host : str
            Name of the host that the EnSight gRPC service is running on.
            The default is ``"127.0.0.1"``, which is the localhost.
        security_code : str
            Shared security code for validating the gRPC communication.
            The default is ``""``.
        verbose : int
            The verbosity level.  If set to 1 or higher the class will call logging.info
            for log output.  The default is ``0``.
        normalize_geometry : bool
            If True, the scene coordinates will be remapped into the volume [-1,-1,-1] - [1,1,1]
            The default is not to remap coordinates.
        vrmode : bool
            If True, do not include the EnSight camera in the generated view group. The default
            is to include the EnSight view in the scene transformations.
        handler : UpdateHandler
            This is an UpdateHandler subclass that is called back when the state of
            a scene transfer changes.  For example, methods are called when the
            transfer begins or ends and when a Part (mesh block) is ready for processing.
        """
        super().__init__()
        self._grpc = ensight_grpc.EnSightGRPC(port=port, host=host, secret_key=security_code)
        self._callback_handler = handler
        self._verbose = verbose
        self._thread: Optional[threading.Thread] = None
        self._message_queue: queue.Queue = queue.Queue()  # Messages coming from EnSight
        self._dsg_queue: Optional[queue.SimpleQueue] = None  # Outgoing messages to EnSight
        self._shutdown = False
        self._dsg = None
        self._normalize_geometry = normalize_geometry
        self._vrmode = vrmode
        self._mesh_block_count = 0
        self._variables: Dict[int, Any] = dict()
        self._groups: Dict[int, Any] = dict()
        self._part: Part = Part(self)
        self._scene_bounds: Optional[List] = None
        self._callback_handler.session = self

    @property
    def scene_bounds(self) -> Optional[List]:
        return self._scene_bounds

    @property
    def mesh_block_count(self) -> int:
        return self._mesh_block_count

    @property
    def vrmode(self) -> bool:
        return self._vrmode

    @vrmode.setter
    def vrmode(self, value: bool) -> None:
        self._vrmode = value

    @property
    def normalize_geometry(self) -> bool:
        return self._normalize_geometry

    @normalize_geometry.setter
    def normalize_geometry(self, value: bool) -> None:
        self._normalize_geometry = value

    @property
    def variables(self) -> dict:
        return self._variables

    @property
    def groups(self) -> dict:
        return self._groups

    @property
    def part(self) -> Part:
        return self._part

    @property
    def grpc(self) -> ensight_grpc.EnSightGRPC:
        return self._grpc

    def log(self, s: str, level: int = 0) -> None:
        """Log a string to the logging system

        If the message level is less than the current verbosity,
        emit the message.
        """
        if level < self._verbose:
            logging.info(s)

    def start(self) -> int:
        """Start a gRPC connection to an EnSight instance

        Make a gRPC connection and start a DSG stream handler.

        Returns
        -------
            0 on success, -1 on an error.
        """
        # Start by setting up and verifying the connection
        self._grpc.connect()
        if not self._grpc.is_connected():
            self.log(f"Unable to establish gRPC connection to: {self._grpc.host}:{self._grpc.port}")
            return -1
        # Streaming API requires an iterator, so we make one from a queue
        # it also returns an iterator.  self._dsg_queue is the input stream interface
        # self._dsg is the returned stream iterator.
        if self._dsg is not None:
            return 0
        self._dsg_queue = queue.SimpleQueue()
        self._dsg = self._grpc.dynamic_scene_graph_stream(
            iter(self._dsg_queue.get, None)  # type:ignore
        )
        self._thread = threading.Thread(target=self._poll_messages)
        if self._thread is not None:
            self._thread.start()
        self._callback_handler.start_connection()
        return 0

    def end(self):
        """Stop a gRPC connection to the EnSight instance"""
        self._callback_handler.end_connection()
        self._grpc.stop_server()
        self._shutdown = True
        self._thread.join()
        self._grpc.shutdown()
        self._dsg = None
        self._thread = None
        self._dsg_queue = None

    def is_shutdown(self):
        """Check the service shutdown request status"""
        return self._shutdown

    def request_an_update(self, animation: bool = False, allow_spontaneous: bool = True) -> None:
        """Start a DSG update
        Send a command to the DSG protocol to "init" an update.

        Parameters
        ----------
        animation:
            if True, export all EnSight timesteps.
        allow_spontaneous:
            if True, allow EnSight to trigger async updates.
        """
        # Send an INIT command to trigger a stream of update packets
        cmd = dynamic_scene_graph_pb2.SceneClientCommand()
        cmd.command_type = dynamic_scene_graph_pb2.SceneClientCommand.INIT
        # Allow EnSight push commands, but full scene only for now...
        cmd.init.allow_spontaneous = allow_spontaneous
        cmd.init.include_temporal_geometry = animation
        cmd.init.allow_incremental_updates = False
        cmd.init.maximum_chunk_size = 1024 * 1024
        self._dsg_queue.put(cmd)  # type:ignore
        # Handle the update messages
        self.handle_one_update()

    def _poll_messages(self) -> None:
        """Core interface to grab DSG events from gRPC and queue them for processing

        This is run by a thread that is monitoring the dsg RPC call for update messages
        it places them in _message_queue as it finds them.  They are picked up by the
        main thread via get_next_message()
        """
        while not self._shutdown:
            try:
                self._message_queue.put(next(self._dsg))  # type:ignore
            except Exception:
                self._shutdown = True
                self.log("DSG connection broken, calling exit")
                os._exit(0)

    def _get_next_message(self, wait: bool = True) -> Any:
        """Get the next queued up protobuffer message

        Called by the main thread to get any messages that were pulled in from the
        dsg stream and placed here by _poll_messages()
        """
        try:
            return self._message_queue.get(block=wait)
        except queue.Empty:
            return None

    def handle_one_update(self) -> None:
        """Monitor the DSG stream and handle a single update operation

        Wait until we get the scene update begin message.  From there, reset the current
        scene buckets and then parse all the incoming commands until we get the scene
        update end command.   At which point, save the generated stage (started in the
        view command handler).  Note: Parts are handled with an available bucket at all times.
        When a new part update comes in or the scene update end happens, the part is "finished".
        """
        # An update starts with a UPDATE_SCENE_BEGIN command
        cmd = self._get_next_message()
        while (cmd is not None) and (
            cmd.command_type != dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_SCENE_BEGIN
        ):
            # Look for a begin command
            cmd = self._get_next_message()

        # Start anew
        self._variables = {}
        self._groups = {}
        self._part = Part(self)
        self._scene_bounds = None
        self._mesh_block_count = 0  # reset when a new group shows up
        self._callback_handler.begin_update()

        # handle the various commands until UPDATE_SCENE_END
        cmd = self._get_next_message()
        while (cmd is not None) and (
            cmd.command_type != dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_SCENE_END
        ):
            self._handle_update_command(cmd)
            cmd = self._get_next_message()

        # Flush the last part
        self._finish_part()

        self._callback_handler.end_update()

    def _handle_update_command(self, cmd: dynamic_scene_graph_pb2.SceneUpdateCommand) -> None:
        """Dispatch out a scene update command to the proper handler

        Given a command object, pull out the correct portion of the protobuffer union and
        pass it to the appropriate handler.

        Parameters
        ----------
        cmd:
            The command to be dispatched.
        """
        name = "Unknown"
        if cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.DELETE_ID:
            name = "Delete IDs"
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_PART:
            name = "Part update"
            tmp = cmd.update_part
            self._handle_part(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_GROUP:
            name = "Group update"
            tmp = cmd.update_group
            self._handle_group(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_GEOM:
            name = "Geom update"
            tmp = cmd.update_geom
            self._part.update_geom(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_VARIABLE:
            name = "Variable update"
            tmp = cmd.update_variable
            self._handle_variable(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_VIEW:
            name = "View update"
            tmp = cmd.update_view
            self._handle_view(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_TEXTURE:
            name = "Texture update"
        self.log(f"{name} --------------------------")

    def _finish_part(self) -> None:
        """Complete the current part

        There is always a part being modified.  This method completes the current part, committing
        it to the handler.
        """
        self._part.build()
        self._callback_handler.finalize_part(self.part)
        self._mesh_block_count += 1

    def _handle_part(self, part: Any) -> None:
        """Handle a DSG UPDATE_PART command

        Finish the current part and set up the next part.

        Parameters
        ----------
        part:
            The command coming from the EnSight stream.
        """
        self._finish_part()
        self._part.reset(part)

    def _handle_group(self, group: Any) -> None:
        """Handle a DSG UPDATE_GROUP command

        Parameters
        ----------
        group:
            The command coming from the EnSight stream.
        """
        # reset current mesh (part) count for unique "part" naming in USD
        self._mesh_block_count = 0

        # record the scene bounds in case they are needed later
        self._groups[group.id] = group
        bounds = group.attributes.get("ENS_SCENE_BOUNDS", None)
        if bounds:
            minmax = list()
            for v in bounds.split(","):
                try:
                    minmax.append(float(v))
                except ValueError:
                    pass
            if len(minmax) == 6:
                self._scene_bounds = minmax
        # callback
        self._callback_handler.add_group(group.id)

    def _handle_variable(self, var: Any) -> None:
        """Handle a DSG UPDATE_VARIABLE command

        Save off the EnSight variable DSG command object.

        Parameters
        ----------
        var:
            The command coming from the EnSight stream.
        """
        self._variables[var.id] = var
        self._callback_handler.add_variable(var.id)

    def _handle_view(self, view: Any) -> None:
        """Handle a DSG UPDATE_VIEW command

        Parameters
        ----------s
        view:
            The command coming from the EnSight stream.
        """
        self._scene_bounds = None
        self._groups[view.id] = view
        self._callback_handler.add_group(view.id, view=True)



#
# This file borrows heavily from the Omniverse Example Connector which
# contains the following notice:
#
###############################################################################
# Copyright 2020 NVIDIA Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
###############################################################################

import argparse
import logging
import math
import os
import shutil
import sys
from typing import Any, Dict, List, Optional

import omni.client
import png
from pxr import Gf, Sdf, Usd, UsdGeom, UsdLux, UsdShade

sys.path.append(os.path.dirname(__file__))
from dsg_server import DSGSession, Part, UpdateHandler  # noqa: E402


class OmniverseWrapper:
    verbose = 0

    @staticmethod
    def logCallback(threadName: None, component: Any, level: Any, message: str) -> None:
        """
        The logger method registered to handle async messages from Omniverse

        If running in verbose mode, reroute the messages to Python Logging.
        """
        if OmniverseWrapper.verbose:
            logging.info(message)

    @staticmethod
    def connectionStatusCallback(
        url: Any, connectionStatus: "omni.client.ConnectionStatus"
    ) -> None:
        """
        If no connection to Omniverse can be made, shut down the service.
        """
        if connectionStatus is omni.client.ConnectionStatus.CONNECT_ERROR:
            sys.exit("[ERROR] Failed connection, exiting.")

    def __init__(
        self,
        live_edit: bool = False,
        path: str = "omniverse://localhost/Users/test",
        verbose: int = 0,
    ) -> None:
        self._cleaned_index = 0
        self._cleaned_names: dict = {}
        self._connectionStatusSubscription = None
        self._stage = None
        self._destinationPath: str = path
        self._old_stages: list = []
        self._stagename = "dsg_scene.usd"
        self._live_edit: bool = live_edit
        if self._live_edit:
            self._stagename = "dsg_scene.live"
        OmniverseWrapper.verbose = verbose

        omni.client.set_log_callback(OmniverseWrapper.logCallback)
        if verbose > 1:
            omni.client.set_log_level(omni.client.LogLevel.DEBUG)

        if not omni.client.initialize():
            sys.exit("[ERROR] Unable to initialize Omniverse client, exiting.")

        self._connectionStatusSubscription = omni.client.register_connection_status_callback(
            OmniverseWrapper.connectionStatusCallback
        )

        if not self.isValidOmniUrl(self._destinationPath):
            self.log("Note technically the Omniverse URL {self._destinationPath} is not valid")

    def log(self, msg: str) -> None:
        """
        Local method to dispatch to whatever logging system has been enabled.
        """
        if OmniverseWrapper.verbose:
            logging.info(msg)

    def shutdown(self) -> None:
        """
        Shutdown the connection to Omniverse cleanly.
        """
        omni.client.live_wait_for_pending_updates()
        self._connectionStatusSubscription = None
        omni.client.shutdown()

    @staticmethod
    def isValidOmniUrl(url: str) -> bool:
        omniURL = omni.client.break_url(url)
        if omniURL.scheme == "omniverse" or omniURL.scheme == "omni":
            return True
        return False

    def stage_url(self, name: Optional[str] = None) -> str:
        """
        For a given object name, create the URL for the item.
        Parameters
        ----------
        name: the name of the object to generate the URL for. If None, it will be the URL for the
              stage name.

        Returns
        -------
        The URL for the object.
        """
        if name is None:
            name = self._stagename
        return self._destinationPath + "/" + name

    def delete_old_stages(self) -> None:
        """
        Remove all the stages included in the "_old_stages" list.
        """
        while self._old_stages:
            stage = self._old_stages.pop()
            omni.client.delete(stage)

    def create_new_stage(self) -> None:
        """
        Create a new stage. using the current stage name.
        """
        self.log(f"Creating Omniverse stage: {self.stage_url()}")
        if self._stage:
            self._stage.Unload()
            self._stage = None
        self.delete_old_stages()
        self._stage = Usd.Stage.CreateNew(self.stage_url())
        # record the stage in the "_old_stages" list.
        self._old_stages.append(self.stage_url())
        UsdGeom.SetStageUpAxis(self._stage, UsdGeom.Tokens.y)
        # in M
        UsdGeom.SetStageMetersPerUnit(self._stage, 1.0)
        self.log(f"Created stage: {self.stage_url()}")

    def save_stage(self, comment: str = "") -> None:
        """
        For live connections, save the current edit and allow live processing.

        Presently, live connections are disabled.
        """
        self._stage.GetRootLayer().Save()  # type:ignore
        omni.client.live_process()

    # This function will add a commented checkpoint to a file on Nucleus if
    # the Nucleus server supports checkpoints
    def checkpoint(self, comment: str = "") -> None:
        """
        Add a checkpoint to the current stage.

        Parameters
        ----------
        comment: str
            If not empty, the comment to be added to the stage
        """
        if not comment:
            return
        result, serverInfo = omni.client.get_server_info(self.stage_url())
        if result and serverInfo and serverInfo.checkpoints_enabled:
            bForceCheckpoint = True
            self.log(f"Adding checkpoint comment <{comment}> to stage <{self.stage_url()}>")
            omni.client.create_checkpoint(self.stage_url(), comment, bForceCheckpoint)

    def username(self, display: bool = True) -> Optional[str]:
        """
        Get the username of the current user.

        Parameters
        ----------
        display : bool, optional if True, send the username to the logging system.

        Returns
        -------
        The username or None.
        """
        result, serverInfo = omni.client.get_server_info(self.stage_url())
        if serverInfo:
            if display:
                self.log(f"Connected username:{serverInfo.username}")
            return serverInfo.username
        return None

    def clear_cleaned_names(self) -> None:
        """
        Clear the list of cleaned names
        """
        self._cleaned_names = {}
        self._cleaned_index = 0

    def clean_name(self, name: str, id_name: Any = None) -> str:
        """Generate a valid USD name

        From a base (EnSight) varname, partname, etc. and the DSG id, generate
        a unique, valid USD name.  Save the names so that if the same name
        comes in again, the previously computed name is returned and if the
        manipulation results in a conflict, the name can be made unique.

        Parameters
        ----------
        name:
            The name to generate a USD name for.

        id_name:
            The DSG id associated with the DSG name, if any.

        Returns
        -------
            A unique USD name.
        """
        # return any previously generated name
        if (name, id_name) in self._cleaned_names:
            return self._cleaned_names[(name, id_name)]
        # replace invalid characters.  EnSight uses a number of characters that are illegal in USD names.
        name = name.replace("+", "_").replace("-", "_")
        name = name.replace(".", "_").replace(":", "_")
        name = name.replace("[", "_").replace("]", "_")
        name = name.replace("(", "_").replace(")", "_")
        name = name.replace("<", "_").replace(">", "_")
        name = name.replace("/", "_").replace("=", "_")
        name = name.replace(",", "_").replace(" ", "_")
        name = name.replace("\\", "_")
        if id_name is not None:
            name = name + "_" + str(id_name)
        if name in self._cleaned_names.values():
            # Make the name unique
            while f"{name}_{self._cleaned_index}" in self._cleaned_names.values():
                self._cleaned_index += 1
            name = f"{name}_{self._cleaned_index}"
        # store off the cleaned name
        self._cleaned_names[(name, id_name)] = name
        return name

    @staticmethod
    def decompose_matrix(values: Any) -> Any:
        """
        Decompose an array of floats (representing a 4x4 matrix) into scale, rotation and translation.
        Parameters
        ----------
        values:
            16 values (input to Gf.Matrix4f CTOR)

        Returns
        -------
        (scale, rotation, translation)
        """
        # ang_convert = 180.0/math.pi
        ang_convert = 1.0
        trans_convert = 1.0
        m = Gf.Matrix4f(*values)
        m = m.GetTranspose()

        s = math.sqrt(m[0][0] * m[0][0] + m[0][1] * m[0][1] + m[0][2] * m[0][2])
        # cleanup scale
        m = m.RemoveScaleShear()
        # r = m.ExtractRotation()
        R = m.ExtractRotationMatrix()
        r = [
            math.atan2(R[2][1], R[2][2]) * ang_convert,
            math.atan2(-R[2][0], 1.0) * ang_convert,
            math.atan2(R[1][0], R[0][0]) * ang_convert,
        ]
        t = m.ExtractTranslation()
        t = [t[0] * trans_convert, t[1] * trans_convert, t[2] * trans_convert]
        return s, r, t

    def create_dsg_mesh_block(
        self,
        name,
        id,
        parent_prim,
        verts,
        conn,
        normals,
        tcoords,
        matrix=[1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        diffuse=[1.0, 1.0, 1.0, 1.0],
        variable=None,
    ):
        # 1D texture map for variables https://graphics.pixar.com/usd/release/tut_simple_shading.html
        # create the part usd object
        partname = self.clean_name(name, id)
        stage_name = "/Parts/" + partname + ".usd"
        part_stage_url = self.stage_url(stage_name)
        omni.client.delete(part_stage_url)
        part_stage = Usd.Stage.CreateNew(part_stage_url)
        self._old_stages.append(part_stage_url)
        xform = UsdGeom.Xform.Define(part_stage, "/" + partname)
        mesh = UsdGeom.Mesh.Define(part_stage, "/" + partname + "/Mesh")
        # mesh.CreateDisplayColorAttr()
        mesh.CreateDoubleSidedAttr().Set(True)
        mesh.CreatePointsAttr(verts)
        mesh.CreateNormalsAttr(normals)
        mesh.CreateFaceVertexCountsAttr([3] * int(conn.size / 3))
        mesh.CreateFaceVertexIndicesAttr(conn)
        if (tcoords is not None) and variable:
            # USD 22.08 changed the primvar API
            if hasattr(mesh, "CreatePrimvar"):
                texCoords = mesh.CreatePrimvar(
                    "st", Sdf.ValueTypeNames.TexCoord2fArray, UsdGeom.Tokens.varying
                )
            else:
                primvarsAPI = UsdGeom.PrimvarsAPI(mesh)
                texCoords = primvarsAPI.CreatePrimvar(
                    "st", Sdf.ValueTypeNames.TexCoord2fArray, UsdGeom.Tokens.varying
                )
            texCoords.Set(tcoords)
            texCoords.SetInterpolation("vertex")
        # sphere = part_stage.DefinePrim('/' + partname + '/sphere', 'Sphere')
        part_prim = part_stage.GetPrimAtPath("/" + partname)
        part_stage.SetDefaultPrim(part_prim)

        # Currently, this will never happen, but it is a setup for rigid body transforms
        # At present, the group transforms have been cooked into the vertices so this is not needed
        matrixOp = xform.AddXformOp(UsdGeom.XformOp.TypeTransform, UsdGeom.XformOp.PrecisionDouble)
        matrixOp.Set(Gf.Matrix4d(*matrix).GetTranspose())

        self.create_dsg_material(
            part_stage, mesh, "/" + partname, diffuse=diffuse, variable=variable
        )
        part_stage.GetRootLayer().Save()

        # glue it into our stage
        path = parent_prim.GetPath().AppendChild("part_ref_" + partname)
        part_ref = self._stage.OverridePrim(path)
        part_ref.GetReferences().AddReference("." + stage_name)

        return part_stage_url

    def create_dsg_material(
        self, stage, mesh, root_name, diffuse=[1.0, 1.0, 1.0, 1.0], variable=None
    ):
        # https://graphics.pixar.com/usd/release/spec_usdpreviewsurface.html
        material = UsdShade.Material.Define(stage, root_name + "/Material")
        pbrShader = UsdShade.Shader.Define(stage, root_name + "/Material/PBRShader")
        pbrShader.CreateIdAttr("UsdPreviewSurface")
        pbrShader.CreateInput("roughness", Sdf.ValueTypeNames.Float).Set(1.0)
        pbrShader.CreateInput("metallic", Sdf.ValueTypeNames.Float).Set(0.0)
        pbrShader.CreateInput("opacity", Sdf.ValueTypeNames.Float).Set(diffuse[3])
        pbrShader.CreateInput("useSpecularWorkflow", Sdf.ValueTypeNames.Int).Set(1)
        if variable:
            stReader = UsdShade.Shader.Define(stage, root_name + "/Material/stReader")
            stReader.CreateIdAttr("UsdPrimvarReader_float2")
            diffuseTextureSampler = UsdShade.Shader.Define(
                stage, root_name + "/Material/diffuseTexture"
            )
            diffuseTextureSampler.CreateIdAttr("UsdUVTexture")
            name = self.clean_name(variable.name)
            filename = self._destinationPath + f"/Parts/Textures/palette_{name}.png"
            diffuseTextureSampler.CreateInput("file", Sdf.ValueTypeNames.Asset).Set(filename)
            diffuseTextureSampler.CreateInput("st", Sdf.ValueTypeNames.Float2).ConnectToSource(
                stReader.ConnectableAPI(), "result"
            )
            diffuseTextureSampler.CreateOutput("rgb", Sdf.ValueTypeNames.Float3)
            pbrShader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).ConnectToSource(
                diffuseTextureSampler.ConnectableAPI(), "rgb"
            )
            stInput = material.CreateInput("frame:stPrimvarName", Sdf.ValueTypeNames.Token)
            stInput.Set("st")
            stReader.CreateInput("varname", Sdf.ValueTypeNames.Token).ConnectToSource(stInput)
        else:
            scale = 1.0
            color = Gf.Vec3f(diffuse[0] * scale, diffuse[1] * scale, diffuse[2] * scale)
            pbrShader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).Set(color)

        material.CreateSurfaceOutput().ConnectToSource(pbrShader.ConnectableAPI(), "surface")
        UsdShade.MaterialBindingAPI(mesh).Bind(material)

        return material

    def create_dsg_variable_textures(self, variables):
        # make folder:   scratch/Textures/{palette_*.png}
        shutil.rmtree("scratch", ignore_errors=True, onerror=None)
        os.makedirs("scratch/Textures", exist_ok=True)
        for var in variables.values():
            data = bytearray(var.texture)
            n_pixels = int(len(data) / 4)
            row = []
            for i in range(n_pixels):
                row.append(data[i * 4 + 0])
                row.append(data[i * 4 + 1])
                row.append(data[i * 4 + 2])
            io = png.Writer(width=n_pixels, height=2, bitdepth=8, greyscale=False)
            rows = [row, row]
            name = self.clean_name(var.name)
            with open(f"scratch/Textures/palette_{name}.png", "wb") as fp:
                io.write(fp, rows)
        uriPath = self._destinationPath + "/Parts/Textures"
        omni.client.delete(uriPath)
        omni.client.copy("scratch/Textures", uriPath)

    def create_dsg_root(self, camera=None):
        root_name = "/Root"
        root_prim = UsdGeom.Xform.Define(self._stage, root_name)
        # Define the defaultPrim as the /Root prim
        root_prim = self._stage.GetPrimAtPath(root_name)
        self._stage.SetDefaultPrim(root_prim)

        if camera is not None:
            cam_name = "/Root/Cam"
            cam_prim = UsdGeom.Xform.Define(self._stage, cam_name)
            cam_pos = Gf.Vec3d(camera.lookfrom[0], camera.lookfrom[1], camera.lookfrom[2])
            target_pos = Gf.Vec3d(camera.lookat[0], camera.lookat[1], camera.lookat[2])
            up_vec = Gf.Vec3d(camera.upvector[0], camera.upvector[1], camera.upvector[2])
            cam_prim = self._stage.GetPrimAtPath(cam_name)
            geom_cam = UsdGeom.Camera(cam_prim)
            if not geom_cam:
                geom_cam = UsdGeom.Camera.Define(self._stage, cam_name)
            # Set camera values
            # center of interest attribute unique for Kit defines the pivot for tumbling the camera
            # Set as an attribute on the prim
            coi_attr = cam_prim.GetAttribute("omni:kit:centerOfInterest")
            if not coi_attr.IsValid():
                coi_attr = cam_prim.CreateAttribute(
                    "omni:kit:centerOfInterest", Sdf.ValueTypeNames.Vector3d
                )
            coi_attr.Set(target_pos)
            # get the camera
            cam = geom_cam.GetCamera()
            # LOL, not sure why is might be correct, but so far it seems to work???
            cam.focalLength = camera.fieldofview
            cam.clippingRange = Gf.Range1f(0.1, 10)
            look_at = Gf.Matrix4d()
            look_at.SetLookAt(cam_pos, target_pos, up_vec)
            trans_row = look_at.GetRow(3)
            trans_row = Gf.Vec4d(-trans_row[0], -trans_row[1], -trans_row[2], trans_row[3])
            look_at.SetRow(3, trans_row)
            # print(look_at)
            cam.transform = look_at

            # set the updated camera
            geom_cam.SetFromCamera(cam)
        return root_prim

    def create_dsg_group(
        self,
        name: str,
        parent_prim,
        obj_type: Any = None,
        matrix: List[float] = [
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ],
    ):
        path = parent_prim.GetPath().AppendChild(self.clean_name(name))
        group_prim = UsdGeom.Xform.Define(self._stage, path)
        # At present, the group transforms have been cooked into the vertices so this is not needed
        matrixOp = group_prim.AddXformOp(
            UsdGeom.XformOp.TypeTransform, UsdGeom.XformOp.PrecisionDouble
        )
        matrixOp.Set(Gf.Matrix4d(*matrix).GetTranspose())
        self.log(f"Created group:'{name}' {str(obj_type)}")
        return group_prim

    def uploadMaterial(self):
        uriPath = self._destinationPath + "/Materials"
        omni.client.delete(uriPath)
        omni.client.copy("resources/Materials", uriPath)

    def createMaterial(self, mesh):
        # Create a material instance for this in USD
        materialName = "Fieldstone"
        newMat = UsdShade.Material.Define(self._stage, "/Root/Looks/Fieldstone")

        matPath = "/Root/Looks/Fieldstone"

        # MDL Shader
        # Create the MDL shader
        mdlShader = UsdShade.Shader.Define(self._stage, matPath + "/Fieldstone")
        mdlShader.CreateIdAttr("mdlMaterial")

        mdlShaderModule = "./Materials/Fieldstone.mdl"
        mdlShader.SetSourceAsset(mdlShaderModule, "mdl")
        # mdlShader.GetPrim().CreateAttribute("info:mdl:sourceAsset:subIdentifier",
        #                                    Sdf.ValueTypeNames.Token, True).Set(materialName)
        # mdlOutput = newMat.CreateSurfaceOutput("mdl")
        # mdlOutput.ConnectToSource(mdlShader, "out")
        mdlShader.SetSourceAssetSubIdentifier(materialName, "mdl")
        shaderOutput = mdlShader.CreateOutput("out", Sdf.ValueTypeNames.Token)
        shaderOutput.SetRenderType("material")
        newMat.CreateSurfaceOutput("mdl").ConnectToSource(shaderOutput)
        newMat.CreateDisplacementOutput("mdl").ConnectToSource(shaderOutput)
        newMat.CreateVolumeOutput("mdl").ConnectToSource(shaderOutput)

        # USD Preview Surface Shaders

        # Create the "USD Primvar reader for float2" shader
        primStShader = UsdShade.Shader.Define(self._stage, matPath + "/PrimST")
        primStShader.CreateIdAttr("UsdPrimvarReader_float2")
        primStShader.CreateOutput("result", Sdf.ValueTypeNames.Float2)
        primStShader.CreateInput("varname", Sdf.ValueTypeNames.Token).Set("st")

        # Create the "Diffuse Color Tex" shader
        diffuseColorShader = UsdShade.Shader.Define(self._stage, matPath + "/DiffuseColorTex")
        diffuseColorShader.CreateIdAttr("UsdUVTexture")
        texInput = diffuseColorShader.CreateInput("file", Sdf.ValueTypeNames.Asset)
        texInput.Set("./Materials/Fieldstone/Fieldstone_BaseColor.png")
        texInput.GetAttr().SetColorSpace("RGB")
        diffuseColorShader.CreateInput("st", Sdf.ValueTypeNames.Float2).ConnectToSource(
            primStShader.CreateOutput("result", Sdf.ValueTypeNames.Float2)
        )
        diffuseColorShaderOutput = diffuseColorShader.CreateOutput("rgb", Sdf.ValueTypeNames.Float3)

        # Create the "Normal Tex" shader
        normalShader = UsdShade.Shader.Define(self._stage, matPath + "/NormalTex")
        normalShader.CreateIdAttr("UsdUVTexture")
        normalTexInput = normalShader.CreateInput("file", Sdf.ValueTypeNames.Asset)
        normalTexInput.Set("./Materials/Fieldstone/Fieldstone_N.png")
        normalTexInput.GetAttr().SetColorSpace("RAW")
        normalShader.CreateInput("st", Sdf.ValueTypeNames.Float2).ConnectToSource(
            primStShader.CreateOutput("result", Sdf.ValueTypeNames.Float2)
        )
        normalShaderOutput = normalShader.CreateOutput("rgb", Sdf.ValueTypeNames.Float3)

        # Create the USD Preview Surface shader
        usdPreviewSurfaceShader = UsdShade.Shader.Define(self._stage, matPath + "/PreviewSurface")
        usdPreviewSurfaceShader.CreateIdAttr("UsdPreviewSurface")
        diffuseColorInput = usdPreviewSurfaceShader.CreateInput(
            "diffuseColor", Sdf.ValueTypeNames.Color3f
        )
        diffuseColorInput.ConnectToSource(diffuseColorShaderOutput)
        normalInput = usdPreviewSurfaceShader.CreateInput("normal", Sdf.ValueTypeNames.Normal3f)
        normalInput.ConnectToSource(normalShaderOutput)

        # Set the linkage between material and USD Preview surface shader
        # usdPreviewSurfaceOutput = newMat.CreateSurfaceOutput()
        # usdPreviewSurfaceOutput.ConnectToSource(usdPreviewSurfaceShader, "surface")
        # UsdShade.MaterialBindingAPI(mesh).Bind(newMat)

        usdPreviewSurfaceShaderOutput = usdPreviewSurfaceShader.CreateOutput(
            "surface", Sdf.ValueTypeNames.Token
        )
        usdPreviewSurfaceShaderOutput.SetRenderType("material")
        newMat.CreateSurfaceOutput().ConnectToSource(usdPreviewSurfaceShaderOutput)

        UsdShade.MaterialBindingAPI.Apply(mesh.GetPrim()).Bind(newMat)

    # Create a distant light in the scene.
    def createDistantLight(self):
        newLight = UsdLux.DistantLight.Define(self._stage, "/Root/DistantLight")
        newLight.CreateAngleAttr(0.53)
        newLight.CreateColorAttr(Gf.Vec3f(1.0, 1.0, 0.745))
        newLight.CreateIntensityAttr(500.0)

    # Create a dome light in the scene.
    def createDomeLight(self, texturePath):
        newLight = UsdLux.DomeLight.Define(self._stage, "/Root/DomeLight")
        newLight.CreateIntensityAttr(2200.0)
        newLight.CreateTextureFileAttr(texturePath)
        newLight.CreateTextureFormatAttr("latlong")

        # Set rotation on domelight
        xForm = newLight
        rotateOp = xForm.AddXformOp(UsdGeom.XformOp.TypeRotateZYX, UsdGeom.XformOp.PrecisionFloat)
        rotateOp.Set(Gf.Vec3f(270, 0, 0))

    def createEmptyFolder(self, emptyFolderPath):
        folder = self._destinationPath + emptyFolderPath
        self.log(f"Creating new folder: {folder}")
        result = omni.client.create_folder(folder)
        self.log(f"Finished creating: {result.name}")
        return result.name


class OmniverseUpdateHandler(UpdateHandler):
    """
    Implement the Omniverse glue to a DSGSession instance
    """

    def __init__(self, omni: OmniverseWrapper):
        super().__init__()
        self._omni = omni
        self._group_prims: Dict[int, Any] = dict()

    def add_group(self, id: int, view: bool = False) -> None:
        super().add_group(id, view)
        group = self.session.groups[id]
        if not view:
            parent_prim = self._group_prims[group.parent_id]
            obj_type = self.get_dsg_cmd_attribute(group, "ENS_OBJ_TYPE")
            matrix = self.group_matrix(group)
            prim = self._omni.create_dsg_group(
                group.name, parent_prim, matrix=matrix, obj_type=obj_type
            )
            self._group_prims[id] = prim
        else:
            # Map a view command into a new Omniverse stage and populate it with materials/lights.
            # Create a new root stage in Omniverse
            self._omni.create_new_stage()
            # Create the root group/camera
            camera_info = group
            if self.session.vrmode:
                camera_info = None
            prim = self._omni.create_dsg_root(camera=camera_info)
            # Create a distance and dome light in the scene
            self._omni.createDomeLight("./Materials/000_sky.exr")
            # Upload a material and textures to the Omniverse server
            self._omni.uploadMaterial()
            self._omni.create_dsg_variable_textures(self.session.variables)
        # record
        self._group_prims[id] = prim

    def add_variable(self, id: int) -> None:
        super().add_variable(id)

    def finalize_part(self, part: Part) -> None:
        # generate an Omniverse compliant mesh from the Part
        command, verts, conn, normals, tcoords, var_cmd = part.build()
        if command is None:
            return
        parent_prim = self._group_prims[command.parent_id]
        obj_id = self._session.mesh_block_count
        matrix = command.matrix4x4
        name = command.name
        color = [
            command.fill_color[0] * command.diffuse,
            command.fill_color[1] * command.diffuse,
            command.fill_color[2] * command.diffuse,
            command.fill_color[3],
        ]
        # Generate the mesh block
        _ = self._omni.create_dsg_mesh_block(
            name,
            obj_id,
            parent_prim,
            verts,
            conn,
            normals,
            tcoords,
            matrix=matrix,
            diffuse=color,
            variable=var_cmd,
        )
        super().finalize_part(part)

    def start_connection(self) -> None:
        super().start_connection()

    def end_connection(self) -> None:
        super().end_connection()

    def begin_update(self) -> None:
        super().begin_update()
        # restart the name tables
        self._omni.clear_cleaned_names()
        # clear the group Omni prims list
        self._group_prims = dict()

    def end_update(self) -> None:
        super().end_update()
        # Stage update complete
        self._omni.save_stage()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Python Omniverse EnSight Dynamic Scene Graph Client",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--path",
        action="store",
        default="omniverse://localhost/Users/test",
        help="Omniverse pathname. Default=omniverse://localhost/Users/test",
    )
    parser.add_argument(
        "--port",
        metavar="ensight_grpc_port",
        nargs="?",
        default=12345,
        type=int,
        help="EnSight gRPC port number",
    )
    parser.add_argument(
        "--host",
        metavar="ensight_grpc_host",
        nargs="?",
        default="127.0.0.1",
        type=str,
        help="EnSight gRPC hostname",
    )
    parser.add_argument(
        "--security",
        metavar="ensight_grpc_security_code",
        nargs="?",
        default="",
        type=str,
        help="EnSight gRPC security code",
    )
    parser.add_argument(
        "--verbose",
        metavar="verbose_level",
        default=0,
        type=int,
        help="Enable debugging information",
    )
    parser.add_argument(
        "--animation", dest="animation", action="store_true", help="Save all timesteps (default)"
    )
    parser.add_argument(
        "--no-animation",
        dest="animation",
        action="store_false",
        help="Save only the current timestep",
    )
    parser.set_defaults(animation=False)
    parser.add_argument(
        "--log_file",
        metavar="log_filename",
        default="",
        type=str,
        help="Save program output to the named log file instead of stdout",
    )
    parser.add_argument(
        "--live",
        dest="live",
        action="store_true",
        default=False,
        help="Enable continuous operation",
    )
    parser.add_argument(
        "--normalize_geometry",
        dest="normalize",
        action="store_true",
        default=False,
        help="Spatially normalize incoming geometry",
    )
    parser.add_argument(
        "--vrmode",
        dest="vrmode",
        action="store_true",
        default=False,
        help="In this mode do not include a camera or the case level matrix.  Geometry only.",
    )
    args = parser.parse_args()

    log_args = dict(format="DSG/Omniverse: %(message)s", level=logging.INFO)
    if args.log_file:
        log_args["filename"] = args.log_file
    logging.basicConfig(**log_args)  # type: ignore

    destinationPath = args.path
    loggingEnabled = args.verbose

    # Build the OmniVerse connection
    target = OmniverseWrapper(path=destinationPath, verbose=loggingEnabled, live_edit=args.live)
    # Print the username for the server
    target.username()

    if loggingEnabled:
        logging.info("Omniverse connection established.")

    # link it to a DSG session
    update_handler = OmniverseUpdateHandler(target)
    dsg_link = DSGSession(
        port=args.port,
        host=args.host,
        vrmode=args.vrmode,
        security_code=args.security,
        verbose=loggingEnabled,
        normalize_geometry=args.normalize,
        handler=update_handler,
    )

    if loggingEnabled:
        dsg_link.log(f"Making DSG connection to: {args.host}:{args.port}")

    # Start the DSG link
    err = dsg_link.start()
    if err < 0:
        sys.exit(err)

    # Simple pull request
    dsg_link.request_an_update(animation=args.animation)

    # Live operation
    if args.live:
        if loggingEnabled:
            dsg_link.log("Waiting for remote push operations")
        while not dsg_link.is_shutdown():
            dsg_link.handle_one_update()

    # Done...
    if loggingEnabled:
        dsg_link.log("Shutting down DSG connection")
    dsg_link.end()

    target.shutdown()
